/* 
 * noIdle v1.1b (date: 28-Feb-2002)
 *
 * Program that don't let user process get more than a specific idle time.
 * 
 * (C) 2002 - Jos Martins (jemart@student.dei.uc.pt)
 * 
 * */

#include "define.h"

static int pipe_fd[2];
static FILE *log_file;

int main(int argc, char *argv[]){
  
  int len, offs=0, pids_offs=0, flag=0;
  char processes[MAX_LINE]="", pids[MAX_LINE]="", user[MAX_LINE]="", token[MAX_LINE]="",pid_token[MAX_LINE]="",log_msg[MAX_LINE];
  struct timeval timeout;
  process c_process,pid_process;
  
  (void)firstAssignments();
  
  printMsg("*************************************************\nnoIdle started");
  
  if(argc==1){
    if(processCmd(CMD_WHOAMI,user,NULL)==EXIT_FAILURE){
      perror("Error processing whoami comand");
      printMsg("Error processing whoami comand\n");
      panicExit(0);
    };
    len=(int)strlen(user);
    user[len-1]='\0';
  }else if(argc==2){
    if(strcmp(argv[1],"--version")==0){
      printf("noIdle version %s (build:%s)\n",VERSION,BUILD);
      panicExit(0);
    };
    strcpy(user,argv[1]);
  }else{
    printf("Wrong usage.\n  usage: %s [USER]",argv[0]);
    panicExit(0);
  };
  
  for(;;){
#ifdef DEBUG
    printMsg("cicle...");
#endif
    
    /* cicle timeout */
    timeout.tv_sec=CICLE_TIME;
    timeout.tv_usec=0;
  
    (void)memset(processes,(int)'\0',(size_t)MAX_LINE);
    /* reads all processes and pids*/
    if(processCmd(CMD_W,user,processes)==EXIT_FAILURE){
      printMsg("Error processing who comand\n");
      panicExit(0);
    };
    
    (void)memset(pids,(int)'\0',(size_t)MAX_LINE);
    if(processCmd(CMD_PS,user,pids)==EXIT_FAILURE){
      printMsg("Error processing ps comand\n");
      panicExit(0);
    };
    
    offs=0;
    /* for each process, check if is idle... */
    for(;;){
      /* catch process information in processes */
      (void)memset(&c_process,(int)'\0',(size_t)sizeof(process));
      (void)memset(token,(int)'\0',(size_t)MAX_LINE);
      
      /* time */
      if(getToken(2,token,processes,&offs)==EXIT_FAILURE){
        perror("Error in get token");
        printMsg("Error in get token #2 from processes\n");
	panicExit(0);
      };
      
      c_process.idle=checkIdle(token);
      
      if(c_process.idle==IDLE){
      
        /* tty */
        if(getToken(1,c_process.tty,processes,&offs)==EXIT_FAILURE){
          perror("Error in get token");
          printMsg("Error in get token #1 from processes\n");
          panicExit(0);
        };
      
        /* name */
        if(getToken(3,c_process.name,processes,&offs)==EXIT_FAILURE){
          perror("Error in get token");
          printMsg("Error in get token #3 from processes\n");
	  panicExit(0);
	};
        
        (void)memset(log_msg,(int)'\0',(size_t)MAX_LINE);
	sprintf(log_msg,"cmd=%s tty=%s time=%s is idle\n",c_process.name,c_process.tty,token);
	printMsg(log_msg);
	
        /* retira o pid e mata */
	pids_offs=0;
	flag=0;
        for(;;){
          (void)memset(&pid_process,(int)'\0',(size_t)sizeof(process));
          (void)memset(pid_token,(int)'\0',(size_t)MAX_LINE);
          /* get name */
	  (void)getToken(2,pid_process.name,pids,&pids_offs);
          /* get tty */
	  (void)getToken(0,pid_process.tty,pids,&pids_offs);
          if( (strcmp(c_process.tty,pid_process.tty)==0) && (strcmp(c_process.name,pid_process.name)==0) ){
            /* get pid */
            (void)getToken(1,pid_token,pids,&pids_offs);
	    c_process.pid=atoi(pid_token);
            (void)memset(log_msg,(int)'\0',(size_t)MAX_LINE);
	    sprintf(log_msg,"  killing cmd=%s tty=%s pid=%d \n",c_process.name,c_process.tty,c_process.pid);
	    printMsg(log_msg);

            /* kill */
	    kill(c_process.pid,SIGINT);
	    kill(c_process.pid,SIGQUIT);
	    kill(c_process.pid,SIGKILL);
	    flag=1;
	    break;
	  };
	  if(flag==1)
            break;
          /* muda de linha */
          while(pids[pids_offs++]!='\n'){};
          if(pids[pids_offs]=='\0')
            break;
	};
	
      };
      
      /* change line */
      while(processes[offs++]!='\n'){};
      if(processes[offs]=='\0')
        break;
    }; /* end of one process. */
    
    (void)select(0,(fd_set*)'\0',(fd_set*)'\0',(fd_set*)'\0',&timeout);
    if(errno==EINTR){
      perror("Ctrl-C pressed:");
      panicExit(0);
    };
  
  };/* end of CICLE_TIME */
  
  return EXIT_SUCCESS;
};

int firstAssignments(){
  if(assignSignals()==EXIT_FAILURE){
    printf("error assigning signals...\n");
    return EXIT_FAILURE;
  };
  
  /* opens pipes */
  if( pipe(pipe_fd)==-1){
    perror("Error making pipe");
    return EXIT_FAILURE;
  };

  /* open file */
  log_file=fopen(LOG_FILE,LOG_PERM);
  if(log_file==NULL){
    perror("Error opening log file");
    panicExit(0);
  };

  return EXIT_SUCCESS;
};

int processCmd(int cmd, char *arg0, char *arg1){
  int child,status;
  
  child=fork();
  if(child==-1){
    /* error */
    perror("Error creating child");
    printMsg("Error creating child\n");
    return EXIT_FAILURE;
  }else if(child!=0){
    /* father */
    if(arg1==NULL)
      (void)read_pipe_fd(arg0);
    else
      (void)read_pipe_fd(arg1);
    
    if( child!=wait(&status) ){
      perror("wrong child!!!");
      printMsg("wrong child!!!\n");
      return EXIT_FAILURE;
    };
    return EXIT_SUCCESS;
  };
  /* child */
  
  if( dup2(pipe_fd[1],STDOUT_FILENO)==-1 ){
    /* error */
    perror("Error creating pipe");
    printMsg("Error creating pipe\n");
    return EXIT_FAILURE;
  };
  
  if( cmd==CMD_WHOAMI )
    (void)execl("/usr/bin/whoami","whoami",NULL);
  if( cmd==CMD_W )
    (void)execl("/usr/bin/w","w","-fsh",arg0,NULL);
  if( cmd==CMD_PS )
    (void)execl("/bin/ps","ps","-U",arg0,"-ho","tty,pid,cmd",NULL);
  perror("Error in processCmd");
  printMsg("Error in processCmd\n");
  return EXIT_FAILURE;
};

int read_pipe_fd(char *str){
  
  int select_val;
  ssize_t count;
  struct timeval rtimeout;
  fd_set fdset;
  
  (void)memset(str,(int)'\0',MAX_LINE);
    
  /* read timeout of 5sec */
  rtimeout.tv_sec=5;
  rtimeout.tv_usec=0;
  
  /* reset fd_set */
  FD_ZERO(&fdset);
  FD_SET(pipe_fd[0],&fdset);
  /* select */
  select_val=select(pipe_fd[0]+1,&fdset,(fd_set*)'\0',(fd_set*)'\0',&rtimeout);
  if(select_val==1){
    count=read(pipe_fd[0],str,(size_t)MAX_LINE);
  }else if(select_val==-1){
    perror("Error in select in main");
    printMsg("Error in select in main\n");
    return EXIT_FAILURE;
  };

  return EXIT_SUCCESS;
};

int checkIdle(char *str){
  int time1=0, time2=0, strLen=0;
#ifdef DEBUG
  char toto[MAX_LINE];
#endif
  char c=' ';
  
  strLen=(int)strlen(str);
  if(sscanf(str,"%d:%d",&time1,&time2)!=2)
    if(sscanf(str,"%d.%d%c",&time1,&time2,&c)!=3)
      return IDLE;
  
  c=str[strLen-1];

#ifdef DEBUG
  (void)sprintf(toto,"got translation t1=%d t2=%d c=%c from %s",time1,time2,c,str);
  printMsg(toto);
#endif
  
  if(isdigit((int)c)!=0){
  /* min:sec */
    if(time1>=IDLE_TIME)
      return IDLE;
    else
      return NOIDLE;
  }else if(c=='s'){
  /* sec:usec */
    return NOIDLE;
  }else if(c=='m'){
  /* h:min */
    return IDLE;
  };
  
  printMsg("is everything else...\n");
  return IDLE;
};

int getToken(int tokNum, char *tok, char *str, int *offs){
  int tokOffs=0, cpOffs=0;
  
  /* get correct token */
  while((tokNum--)!=0){
    while(str[*offs+tokOffs]!=' '){
      tokOffs++;
      if(str[*offs+tokOffs]=='\n')
        return EXIT_FAILURE;
    };
    while(str[*offs+tokOffs]==' '){tokOffs++;};
  };
  
  /* copy token */
  while(str[*offs+tokOffs]!=' '){
    tok[cpOffs++]=str[*offs+tokOffs++];
    if(tok[cpOffs-1]=='\n'){
      tok[cpOffs-1]='\0';
      return EXIT_SUCCESS;
    }
  };
  tok[tokOffs]='\0';
  return EXIT_SUCCESS;
};

void printMsg(char *msg){
 time_t time_mark;
 
 if(log_file!=NULL){
   (void)time(&time_mark);
   fprintf(log_file,"%s",ctime(&time_mark));
   /* pop one char from log_file */
   fprintf(log_file,"  %s\n",msg);
   (void)fflush(log_file);
  };
  return ;
};

int assignSignals(void){
  if (signal(SIGSEGV,panicExit)==SIG_ERR ||
      signal(SIGTERM,panicExit)==SIG_ERR ||
      signal(SIGINT,panicExit)==SIG_ERR)
    return EXIT_FAILURE;
  else
    return EXIT_SUCCESS;
};
 
void panicExit(int signal){
  printf("\n\nGOT PANIC!!!");
  
  if(signal==SIGTERM){
    printf("received SIGTERM");
    printMsg("received SIGTERM. Exiting...");
  }else if(signal==SIGINT){
    printf("received SIGSEGV");
    printMsg("received SIGSEGV. Exiting...");
  }else if(signal==SIGSEGV){
    printf("received SIGSEGV");
    printMsg("received SIGSEGV. Exiting...");
  }else{
    printMsg("exit called...\n");
  };

  /* A-> log_file==null
   * B-> fclose!=0 */
  /* (A+B) == !( (!A) * (!B) ) */
  if(!( (log_file!=NULL) && (fclose(log_file)==0)))
    perror("Error closing log file");
	  
  printf("\nexiting...\n");
  exit(EXIT_SUCCESS);
};
